"""Classes to represent template dependencies.
"""
import jinja2
from typing import Optional, List, Dict


class Target:
    """The target of a dependency.

    A target is dynamic if it is not hard-coded in the template, in which case
    the name of the target is unknown until it is resolved.
    """

    def __init__(self, dynamic: bool, name: Optional[str]):
        """Initialises a new `Target` class.

        This class should not be instantiated manually.
        """
        self.__dynamic = dynamic
        self.__name = name

    def __repr__(self):
        return f"Target(dynamic={self.__dynamic}, name={self.__name})"

    def __eq__(self, other):
        if not isinstance(other, Target):
            return NotImplemented

        return self.__dynamic == other.__dynamic and self.__name == other.__name

    @property
    def is_dynamic(self) -> bool:
        """True if the target name can't be known until the template is
        rendered.
        """
        return self.__dynamic

    @property
    def name(self) -> Optional[str]:
        """The name of the target template, or `None` if the target is dynamic."""
        return self.__name


class _ResolvedTarget:
    def __init__(self, name: str):
        self.__name = name
        self.__last_watch = True

    @property
    def last_watch(self) -> bool:
        return self.__last_watch

    def watch_reset(self):
        self.__last_watch = False

    @property
    def name(self) -> str:
        return self.__name


class Dependency:
    """A dependency to one or more templates."""

    def __init__(
        self,
        dependency_type: str,
        targets: List[Target],
        with_context: Optional[bool] = None,
        ignore_missing: Optional[bool] = None,
        imported_as: Optional[str] = None,
        imported_names: Optional[List[str]] = None,
    ):
        """Initialises a new `Dependency` class.

        This class should not be instantiated manually.
        """
        self.__type = dependency_type
        self.__targets = targets
        self.__with_context = with_context
        self.__ignore_missing = ignore_missing
        self.__imported_as = imported_as
        self.__imported_names = imported_names
        self.__resolved: List[_ResolvedTarget] = []

    def _resolve(self, name: str):
        self.__resolved.append(_ResolvedTarget(name))

    def _watch_reset(self):
        for r in self.__resolved:
            r.watch_reset()

    @property
    def type(self) -> str:
        """The type of dependency. May be one of ``"extends"``, ``"include"`` or
        ``"import"``.
        """
        return self.__type

    @property
    def target(self) -> Optional[Target]:
        """The target of the dependency, if there is only one, or ``None``.

        .. note::
           Use `targets <#jinja2td.Dependency.targets>`_ instead if you
           need to handle multiple targets.
        """
        return self.__targets[0] if len(self.__targets) == 1 else None

    @property
    def targets(self) -> List[Target]:
        """All the targets of the dependency, in order.

        For example, an ``{% include ['template', 'fallback'] %}`` will have two
        targets.
        """
        return self.__targets.copy()

    @property
    def with_context(self) -> Optional[bool]:
        """Whether the context is passed to the dependency or not.

        .. note::
           Only available with ``"include"`` and ``"import"`` dependencies.
        """
        return self.__with_context

    @property
    def ignore_missing(self) -> Optional[bool]:
        """Whether the dependency is optional or not.

        .. note::
           Only available with ``"include"`` dependencies.
        """
        return self.__ignore_missing

    @property
    def imported_as(self) -> Optional[str]:
        """Whether the dependency is optional or not.

        .. note::
           Only available with ``"import"`` dependencies.
        """
        return self.__imported_as

    @property
    def imported_names(self) -> Optional[List[str]]:
        """Whether the dependency is optional or not.

        .. note::
           Only available with ``"import"`` dependencies.
        """
        return self.__imported_names

    @property
    def resolved(self) -> List[str]:
        """The names of the templates actually imported by this dependency.

        .. warning::
           Templates used by async environments aren't taken into account by
           default. See `DependencyGraph.used_last_watch <#jinja2td.DependencyGraph.used_last_watch>`_
           for more information.
        """
        return [r.name for r in self.__resolved]

    @property
    def resolved_last_watch(self) -> List[str]:
        """The names of the templates imported during the last watch.="""
        return [r.name for r in self.__resolved if r.last_watch]


class Template:
    """Represent a template in the dependency graph.

    This is NOT a Jinja2 template.
    """

    def __init__(self, name: str, file: Optional[str], graph: "DependencyGraph"):
        """Initialises a new `Template` class.

        This class should not be instantiated manually.
        """
        self.__name = name
        self.__file = file
        self.__deps: List[Dependency] = []
        self.__modified = False
        self.__graph = graph

    def _set_modified(self):
        self.__modified = False

    def _add_dependency(self, dependency: Dependency) -> int:
        for i, d in enumerate(self.__deps):
            if d == dependency:
                return i  # don't register the same dependency twice
        self.__deps.append(dependency)
        return len(self.__deps) - 1

    def _resolve_dependency(self, dependency_id: int, name: str):
        self.__deps[dependency_id]._resolve(name)

    def _watch_reset(self):
        for d in self.__deps:
            d._watch_reset()

    @property
    def name(self) -> str:
        """The name of the template."""
        return self.__name

    @property
    def file(self) -> Optional[str]:
        """The source file of the template, or ``None`` if the template wasn't
        loaded from a file.
        """
        return self.__file

    @property
    def dependencies(self) -> List[Dependency]:
        """The dependencies of this template."""
        return self.__deps.copy()

    @property
    def was_modified(self) -> bool:
        """`True` if the template was loaded multiple times."""
        return self.__modified

    def get_includes(self) -> List[Dependency]:
        """Get all ``"include"`` dependencies.

        :returns: The list of al ``"include"`` dependencies.
        """
        return [d for d in self.__deps if d.type == "include"]

    def get_imports(self) -> List[Dependency]:
        """Get all ``"import"`` dependencies.

        :returns: The list of all ``"import"`` dependencies.
        """
        return [d for d in self.__deps if d.type == "import"]

    def get_parent(self) -> Optional[Dependency]:
        """Get the parent template, if there is one.

        :returns: An ``"extends"`` dependency or ``None``.
        """
        extends = [d for d in self.__deps if d.type == "extends"]
        if len(extends) == 1:
            return extends[0]
        else:
            return None

    def find_included(self) -> List["Template"]:
        """Get the templates that include this one.

        :returns: The list of templates with an ``"include"`` dependency
                  targeting this template.
        """
        found = []

        for t in self.__graph.templates:
            for d in t.get_includes():
                if Target(False, self.name) in d.targets:
                    found.append(t)

        return found

    def find_imported(self) -> List["Template"]:
        """Get the templates that import this one.

        :returns: The list of templates with an ``"import"`` dependency
                  targeting this template.
        """
        found = []

        for t in self.__graph.templates:
            for d in t.get_imports():
                if d.target == Target(False, self.name):
                    found.append(t)

        return found

    def find_children(self) -> List["Template"]:
        """Get the templates extending this one.

        :returns: The list of templates with an ``"extends"`` dependency
                  targeting this template.
        """
        found = []

        for t in self.__graph.templates:
            parent = t.get_parent()
            if parent is not None:
                if parent.target == Target(False, self.name):
                    found.append(t)

        return found


class DependencyGraph:
    """A collection of templates and their dependencies.

    This is the type of the ``dependencies`` attribute of the environment.
    """

    def __init__(self):
        """Initialises a new `DependencyGraph` class.

        This class should not be instantiated manually.
        """
        self.__templates: Dict[str, Template] = {}
        self.__watch_async = False

    def _add_template(self, name: str, file: Optional[str]):
        if name in self.__templates:
            self.__templates[name]._set_modified()
        else:
            self.__templates[name] = Template(name, file, self)

    def _register_dependency(
        self,
        dependent: str,
        dependency_type: str,
        targets: List[Target],
        **kwargs,
    ) -> int:
        dependency = Dependency(dependency_type, targets, **kwargs)

        if dependent not in self.__templates:
            raise ValueError(f"No such template: {dependent}")

        return self.__templates[dependent]._add_dependency(dependency)

    def _resolve_dependency(
        self,
        dependent: str,
        dependency_id: int,
        template: jinja2.Template,
    ) -> jinja2.Template:
        if dependent in self.__templates and template.name is not None:
            self.__templates[dependent]._resolve_dependency(
                dependency_id, template.name
            )
        # otherwise, ignore silently not to break existing code

        return template

    @property
    def templates(self) -> List[Template]:
        """All the templates known to the environment."""
        return list(self.__templates.values())

    @property
    def watch_async(self) -> bool:
        """See `used_last_watch <#jinja2td.DependencyGraph.used_last_watch>`_."""
        return self.__watch_async

    @watch_async.setter
    def watch_async(self, value: bool):
        self.__watch_async = value

    def get_template(self, name: str) -> Optional[Template]:
        """Get a template.

        :param name: The name of the template, the same you would pass to
                     `jinja2.Environment.get_template <https://jinja.palletsprojects.com/en/3.1.x/api/#jinja2.Environment.get_template>`_.

        :returns: The corresponding template, or None if the template is
                  unknown.
        """
        return self.__templates.get(name)

    def watch(self):
        """Start watching for templates used.

        Call it before rendering a template, and then use
        `used_last_watch <#jinja2td.DependencyGraph.used_last_watch>`_
        to get all the templates used to build it.
        """
        for t in self.__templates.values():
            t._watch_reset()

    def used_last_watch(self) -> List[Template]:
        """Returns all the templates used for rendering templates since the last
        call to `watch`.

        By default, the watch system is disabled in async environments. You can
        enable it explicitly by setting `watch_async <#jinja2td.DependencyGraph.watch_async>`_
        to ``True``, but *be careful not to call* `watch <#jinja2td.DependencyGraph.watch>`_
        *or* `used_last_watch <#jinja2td.DependencyGraph.used_last_watch>`_
        *while a template is rendering*.

        :returns: The names of the templates used during the last watch.
        """
        since_last_watch = []
        for t in self.__templates.values():
            for d in t.dependencies:
                since_last_watch += d.resolved_last_watch
        return [self.__templates[name] for name in set(since_last_watch)]



"""Alter the behavior of the Jinja template compiler. 
"""
from jinja2.compiler import CodeGenerator, Frame, t, CompilerExit
from jinja2 import nodes

from .dependencies import Target


def _override(cls):
    def deco(func):
        docstring = getattr(cls, func.__name__).__doc__
        if docstring is None:
            docstring = ""
        else:
            docstring += "\n\n"
        func.__doc__ = docstring + "(Warning: this method has been altered by jinja2td)"
        setattr(cls, func.__name__, func)

    return deco


@_override(CodeGenerator)
def visit_Include(self, node: nodes.Include, frame: Frame) -> None:
    # The code in this section has been copied verbatim from Jinja2 (file compiler.py, lines 1044 to 1067)
    # https://github.com/pallets/jinja/blob/15206881c006c79667fe5154fe80c01c65410679/src/jinja2/compiler.py#L1044-L1067
    # Copyright 2007 Pallets - This code is licensed under the BSD 3-Clause license.
    # See LICENSE_JINJA2 for the full license text.
    # BEGIN COPIED CODE
    if node.ignore_missing:
        self.writeline("try:")
        self.indent()

    func_name = "get_or_select_template"
    if isinstance(node.template, nodes.Const):
        if isinstance(node.template.value, str):
            func_name = "get_template"
        elif isinstance(node.template.value, (tuple, list)):
            func_name = "select_template"
    elif isinstance(node.template, (nodes.Tuple, nodes.List)):
        func_name = "select_template"

    self.writeline(f"template = environment.{func_name}(", node)
    self.visit(node.template, frame)
    self.write(f", {self.name!r})")
    if node.ignore_missing:
        self.outdent()
        self.writeline("except TemplateNotFound:")
        self.indent()
        self.writeline("pass")
        self.outdent()
        self.writeline("else:")
        self.indent()
    # END COPIED CODE

    if hasattr(self.environment, "dependencies"):
        targets = [Target(True, None)]
        if isinstance(node.template, nodes.Const):
            if isinstance(node.template.value, str):
                targets = [Target(False, node.template.value)]
            elif isinstance(node.template.value, (tuple, list)):
                targets = [Target(False, name) for name in node.template.value]
        elif isinstance(node.template, (nodes.Tuple, nodes.List)):
            targets = [
                Target(False, item.value)
                if (isinstance(item, nodes.Const) and isinstance(item.value, str))
                else Target(True, None)
                for item in node.template.items
            ]

        dependency_id = self.environment.dependencies._register_dependency(
            dependent=self.name,
            dependency_type="include",
            targets=targets,
            with_context=node.with_context,
            ignore_missing=node.ignore_missing,
        )

        # tracking templates in async environments could break the watch system
        if not self.environment.is_async or self.environment.dependencies.watch_async:
            self.writeline("if hasattr(environment, 'dependencies'):")
            self.indent()
            self.writeline(
                f"environment.dependencies._resolve_dependency({self.name!r}, {dependency_id}, template)"
            )
            self.outdent()

    # The code in this section has been copied verbatim from Jinja2 (file compiler.py, lines 1069 to 1098)
    # https://github.com/pallets/jinja/blob/15206881c006c79667fe5154fe80c01c65410679/src/jinja2/compiler.py#L1069-L1098
    # Copyright 2007 Pallets - This code is licensed under the BSD 3-Clause license.
    # See LICENSE_JINJA2 for the full license text.
    # BEGIN COPIED CODE
    def loop_body() -> None:
        self.indent()
        self.simple_write("event", frame)
        self.outdent()

    if node.with_context:
        self.writeline(
            f"gen = template.root_render_func("
            "template.new_context(context.get_all(), True,"
            f" {self.dump_local_context(frame)}))"
        )
        self.writeline("try:")
        self.indent()
        self.writeline(f"{self.choose_async()}for event in gen:")
        loop_body()
        self.outdent()
        self.writeline(
            f"finally: {self.choose_async('await gen.aclose()', 'gen.close()')}"
        )
    elif self.environment.is_async:
        self.writeline(
            "for event in (await template._get_default_module_async())"
            "._body_stream:"
        )
        loop_body()
    else:
        self.writeline("yield from template._get_default_module()._body_stream")

    if node.ignore_missing:
        self.outdent()
    # END COPIED CODE


@_override(CodeGenerator)
def _import_common(
    self,
    node: t.Union[nodes.Import, nodes.FromImport],
    frame: Frame,
) -> None:
    if hasattr(self.environment, "dependencies"):
        targets = [Target(True, None)]
        if isinstance(node.template, nodes.Const):
            if isinstance(node.template.value, str):
                targets = [Target(False, node.template.value)]

        dependency_id = self.environment.dependencies._register_dependency(
            dependent=self.name,
            dependency_type="import",
            targets=targets,
            with_context=node.with_context,
            imported_as=node.target if isinstance(node, nodes.Import) else None,
            imported_names=node.names if isinstance(node, nodes.FromImport) else None,
        )

        # tracking templates in async environments could break the watch system
        if not self.environment.is_async or self.environment.dependencies.watch_async:
            self.write("(lambda template: ")
            self.write(
                f"environment.dependencies._resolve_dependency({self.name!r}, {dependency_id}, template)"
            )
            self.write(" if hasattr(environment, 'dependencies') else template)(")

    # The code in this section has been copied verbatim from Jinja2 (file compiler.py, lines 1103 to 1104)
    # https://github.com/pallets/jinja/blob/15206881c006c79667fe5154fe80c01c65410679/src/jinja2/compiler.py#L1103-L1104
    # Copyright 2007 Pallets - This code is licensed under the BSD 3-Clause license.
    # See LICENSE_JINJA2 for the full license text.
    # BEGIN COPIED CODE
    self.write(f"{self.choose_async('await ')}environment.get_template(")
    self.visit(node.template, frame)
    # END COPIED CODE

    self.write(f", {self.name!r})")
    if hasattr(self.environment, "dependencies"):
        if not self.environment.is_async or self.environment.dependencies.watch_async:
            self.write(")")  # close the parenthesis open at the end of line 152
    self.write(".")

    # The code in this section has been copied verbatim from Jinja2 (file compiler.py, lines 1107 to 1113)
    # https://github.com/pallets/jinja/blob/15206881c006c79667fe5154fe80c01c65410679/src/jinja2/compiler.py#L1107-L1113
    # Copyright 2007 Pallets - This code is licensed under the BSD 3-Clause license.
    # See LICENSE_JINJA2 for the full license text.
    # BEGIN COPIED CODE
    if node.with_context:
        f_name = f"make_module{self.choose_async('_async')}"
        self.write(
            f"{f_name}(context.get_all(), True, {self.dump_local_context(frame)})"
        )
    else:
        self.write(f"_get_default_module{self.choose_async('_async')}(context)")
    # END COPIED CODE


@_override(CodeGenerator)
def visit_Extends(self, node: nodes.Extends, frame: Frame) -> None:
    # The code in this section has been copied verbatim from Jinja2 (file compiler.py, lines 1002 to 1027)
    # https://github.com/pallets/jinja/blob/15206881c006c79667fe5154fe80c01c65410679/src/jinja2/compiler.py#L1002-L1027
    # Copyright 2007 Pallets - This code is licensed under the BSD 3-Clause license.
    # See LICENSE_JINJA2 for the full license text.
    # BEGIN COPIED CODE
    if not frame.toplevel:
        self.fail("cannot use extend from a non top-level scope", node.lineno)

    # if the number of extends statements in general is zero so
    # far, we don't have to add a check if something extended
    # the template before this one.
    if self.extends_so_far > 0:
        # if we have a known extends we just add a template runtime
        # error into the generated code.  We could catch that at compile
        # time too, but i welcome it not to confuse users by throwing the
        # same error at different times just "because we can".
        if not self.has_known_extends:
            self.writeline("if parent_template is not None:")
            self.indent()
        self.writeline('raise TemplateRuntimeError("extended multiple times")')

        # if we have a known extends already we don't need that code here
        # as we know that the template execution will end here.
        if self.has_known_extends:
            raise CompilerExit()
        else:
            self.outdent()

    self.writeline("parent_template = environment.get_template(", node)
    self.visit(node.template, frame)
    self.write(f", {self.name!r})")
    # END COPIED CODE

    if hasattr(self.environment, "dependencies"):
        targets = [Target(True, None)]
        if isinstance(node.template, nodes.Const):
            if isinstance(node.template.value, str):
                targets = [Target(False, node.template.value)]

        dependency_id = self.environment.dependencies._register_dependency(
            dependent=self.name,
            dependency_type="extends",
            targets=targets,
        )

        # tracking templates in async environments could break the watch system
        if not self.environment.is_async or self.environment.dependencies.watch_async:
            self.writeline("if hasattr(environment, 'dependencies'):")
            self.indent()
            self.writeline(
                f"environment.dependencies._resolve_dependency({self.name!r}, {dependency_id}, parent_template)"
            )
            self.outdent()

    # The code in this section has been copied verbatim from Jinja2 (file compiler.py, lines 1028 to 1040)
    # https://github.com/pallets/jinja/blob/15206881c006c79667fe5154fe80c01c65410679/src/jinja2/compiler.py#L1028-L1040
    # Copyright 2007 Pallets - This code is licensed under the BSD 3-Clause license.
    # See LICENSE_JINJA2 for the full license text.
    # BEGIN COPIED CODE
    self.writeline("for name, parent_block in parent_template.blocks.items():")
    self.indent()
    self.writeline("context.blocks.setdefault(name, []).append(parent_block)")
    self.outdent()

    # if this extends statement was in the root level we can take
    # advantage of that information and simplify the generated code
    # in the top level from this point onwards
    if frame.rootlevel:
        self.has_known_extends = True

    # and now we have one more
    self.extends_so_far += 1
    # END COPIED CODE



"""The Jinja2 extension for jinja2-td.
"""

import jinja2
from jinja2.ext import Extension

from . import overrides as _
from .dependencies import DependencyGraph


class Introspection(Extension):
    """An extension that provides access to the features of jinja2td."""

    def __init__(self, environment):
        super().__init__(environment)

        self.__deps = DependencyGraph()

        environment.extend(dependencies=self.__deps)

    def preprocess(self, source, name, filename=None):
        self.__deps._add_template(name, filename)
        return source



from setuptools import setup

with open("README.md", "rt", encoding="utf8") as f:
    README = f.read()

setup(
    name="Jinja2-TD",
    version="3.1.6",
    description="Jinja2 template dependency insight",
    long_description=README,
    long_description_content_type="text/markdown",
    license="MIT",
    author="Louis DEVIE",
    url="https://github.com/louisdevie/jinja2-td",
    packages=["jinja2td"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3.8",
    ],
    keywords=[
        "jinja",
        "jinja2",
        "template",
        "dependency",
    ],
    python_requires=">=3.8",
    install_requires=[
        "Jinja2 >=3.1.5, <=3.1.6",
    ],
)



"""Provides information about Jinja2 template dependencies.

This package will plug itself into the template compiler, and provide an
interface that lets you use it without having to modify any of your existing
code.
"""

import jinja2

__ALLOWED_JINJA_VERSIONS = ("3.1.5", "3.1.6")
if jinja2.__version__ not in __ALLOWED_JINJA_VERSIONS:
    import warnings

    warnings.warn(
        f"Jinja version doesn't match: expected one of {__ALLOWED_JINJA_VERSIONS}, got {jinja2.__version__!r}",
        RuntimeWarning,
    )

from . import overrides as _
from .introspection import Introspection
from .dependencies import DependencyGraph, Template, Dependency, Target
